import unittest
from unittest.mock import patch
from ..src.bookkeeping.analysis.summary import SummaryAnalysis

"""
Note: In summary.py, unit tests cannot be appropriately applied to displayTrend() and displayByLabel()
because of their functionalities in visualization. 
Therefore, no unit test case for these functions is included in this test class, as agreed upon by the instructor.
"""

class TestSummaryAnalysis(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        print("Setting up the test class")

    @classmethod
    def tearDownClass(cls):
        print("Tearing down the test class")

    def setUp(self):
        print("Setting up for a test")
        self.analysis_instance = SummaryAnalysis(file_path = "bookkeeping/userBook.json")

    def tearDown(self):
        self.analysis_instance = None
        print("Tearing down after a test")


    def test_checkBalance(self):
        # Test with original balance
        with patch("builtins.print") as mock_print:
            self.analysis_instance.checkBalance()
            mock_print.assert_called_once_with("Overall Balance for 2023-12: 482.0")

        # Test with a positive balance
        self.analysis_instance.transactions = [
            {'type': 'income', 'amount': 100.0, 'date': '2023-12-01', "desc": "source", "label": "salary income"},
            {'type': 'expense', 'amount': 20.0, 'date': '2023-12-05', "desc": "description", "label": "grocery"}
        ]
        with patch("builtins.print") as mock_print:
            self.analysis_instance.checkBalance()
            mock_print.assert_called_once_with("Overall Balance for 2023-12: 80.0")

        # Test with a zero balance
        self.analysis_instance.transactions = []
        with patch("builtins.print") as mock_print:
            self.analysis_instance.checkBalance()
            mock_print.assert_called_once_with("Overall Balance for 2023-12: 0")

        # Test with a negative balance
        self.analysis_instance.transactions = [
            {'type': 'expense', 'amount': 1000.0, 'date': '2023-12-20', "desc": "description", "label": "grocery"}
        ]
        with patch("builtins.print") as mock_print:
            self.analysis_instance.checkBalance()
            mock_print.assert_called_once_with("Overall Balance for 2023-12: -1000.0")


if __name__ == '__main__':
    unittest.main(argv =[''], verbosity=2, exit=False)


import unittest
from unittest.mock import patch, mock_open
from ..src.bookkeeping.analysis.analytics import Analysis
class TestAnalysis(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        print("Setting up the test class")

    @classmethod
    def tearDownClass(cls):
        print("Tearing down the test class")

    def setUp(self):
        print("Setting up for a test")
        self.analysis_instance = Analysis(file_path = "userBook.json")
        self.income = [t for t in self.analysis_instance.transactions if t['type'] == 'income']
        self.expense = [t for t in self.analysis_instance.transactions if t['type'] == 'expense']

    def tearDown(self):
        self.analysis_instance = None
        print("Tearing down after a test")

    def test_loadBookData_expense(self):
        # Test loading expense transactions
        expected_expense_count = 30
        expected_first_expense_amount = 132.0
        expected_second_expense_label = "entertainment"

        self.assertEqual(len(self.expense), expected_expense_count)
        self.assertAlmostEqual(self.expense[0]['amount'], expected_first_expense_amount, places=2)
        self.assertEqual(self.expense[1]['label'], expected_second_expense_label)
        self.assertEqual(self.expense[1]['type'], 'expense')

    def test_loadBookData_income(self):
        # Test loading income transactions
        expected_income_count = 33
        expected_first_income_date = "2023-04-10"
        expected_second_income_label = "capital gains"

        self.assertEqual(len(self.income), expected_income_count)
        self.assertEqual(self.income[0]['date'], expected_first_income_date)
        self.assertEqual(self.income[1]['label'], expected_second_income_label)
        self.assertEqual(self.income[1]['type'], 'income')


    def test_loadBookData_file_not_found(self):
        # Mock the open function to simulate file not found and invalid JSON
        with patch("builtins.open", side_effect=[FileNotFoundError("File not found"), mock_open(read_data="This is not a valid JSON"), None]):
            with patch("builtins.print") as mock_print:
                # File not found
                self.analysis_instance = Analysis("nonexistent_file.json")
                transactions_file_not_found = self.analysis_instance.transactions     
                self.assertEqual(transactions_file_not_found, [])
                mock_print.assert_called_once_with("Error: File 'nonexistent_file.json' not found.")



if __name__ == '__main__':
    unittest.main(argv =[''], verbosity=2, exit=False)



import unittest
import os
from ..src.bookkeeping.management.transaction import Income, Expense
from ..src.bookkeeping.management.book import Book

import sys
from io import StringIO

class TestBook(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        # Create a Book instance for testing
        cls.filePath = "testJSON.json"
        cls.testBook = Book(cls.filePath)

    def setUp(self):
        self.testBook.clearAll(self.filePath)

    def tearDown(self):
        self.testBook.clearAll(self.filePath)

    @classmethod
    def tearDownClass(cls):
        try:
            os.remove(cls.filePath)
        except OSError:
            pass
        del cls.filePath
        del cls.testBook
        

    def test_addTran_income(self):
        # Test adding an income transaction
        income_transaction = Income(100.0, "Test Income", "salary income")
        self.testBook.addTran(income_transaction, self.filePath)

        self.assertEqual(len(self.testBook.book["income"]), 1)
        self.assertEqual(self.testBook.book["income"][0]["amount"], 100.0)
        self.assertEqual(self.testBook.book["income"][0]["desc"], "Test Income")
        self.assertEqual(self.testBook.book["income"][0]["label"], "salary income")

    def test_addTran_expense(self):
        # Test adding an expense transaction
        expense_transaction = Expense(50.0, "Test Expense", "grocery")
        self.testBook.addTran(expense_transaction, self.filePath)

        self.assertEqual(len(self.testBook.book["expense"]), 1)
        self.assertEqual(self.testBook.book["expense"][0]["amount"], 50.0)
        self.assertEqual(self.testBook.book["expense"][0]["desc"], "Test Expense")
        self.assertEqual(self.testBook.book["expense"][0]["label"], "grocery")

    def test_searchTran_valid(self):

        income_transaction = Income(200.0, "Test Income", "passive income")
        self.testBook.addTran(income_transaction, self.filePath)
        
        # Capture the output
        capturedOutput = StringIO()          
        sys.stdout = capturedOutput                   
        # Search for the transaction
        output = self.testBook.searchTran("income", "Test Income")                     
        sys.stdout = sys.__stdout__
        
        result = capturedOutput.getvalue()

        self.assertTrue(output)
        self.assertIn("income", result)
        self.assertIn("Test Income", result)
        self.assertIn("passive income", result)
        

    def test_searchTran_invalid(self):
        # Test searching for a transaction that does not exist
        income_transaction = Income(200.0, "Searchable Income", "passive income")
        self.testBook.addTran(income_transaction, self.filePath)
        
        # Capture the output
        capturedOutput = StringIO()          
        sys.stdout = capturedOutput                   
        
        output = self.testBook.searchTran("income", "Non-Existent")                     
        sys.stdout = sys.__stdout__
        
        result = capturedOutput.getvalue()

        self.assertFalse(output)
        self.assertIn("Non-Existent", result)
        self.assertIn("not found", result)
        self.assertIn("income", result)
        

    def test_removeTran_valid(self):
       
        income_transaction = Income(300.0, "Removable Income", "salary income")
        self.testBook.addTran(income_transaction, self.filePath)
        
        capturedOutput = StringIO()          
        sys.stdout = capturedOutput                   
        
        self.testBook.removeTran("income", 1, self.filePath)                     
        sys.stdout = sys.__stdout__
        
        result = capturedOutput.getvalue()

        self.assertEqual(len(self.testBook.book["income"]), 0)
        self.assertIn("income", result)
        self.assertIn("Removable Income", result)
        self.assertIn("salary income", result)

    def test_removeTran_invalid_index(self):
       
        income_transaction = Income(300.0, "Removable Income", "salary income")
        self.testBook.addTran(income_transaction, self.filePath)

       
        capturedOutput = StringIO()          
        sys.stdout = capturedOutput                   
        
        self.testBook.removeTran("income", 2, self.filePath)                     
        sys.stdout = sys.__stdout__
        
        result = capturedOutput.getvalue()

        self.assertEqual(len(self.testBook.book["income"]), 1)
        self.assertIn("index is out of bounds", result)

    def test_clearAll(self):
        
        income_transaction = Income(400.0, "Clearable Income", "passive income")
        expense_transaction = Expense(50.0, "Clearable Expense", "grocery")
        self.testBook.addTran(income_transaction, self.filePath)
        self.testBook.addTran(expense_transaction, self.filePath)

        self.testBook.clearAll(self.filePath)
        
        self.assertEqual(len(self.testBook.book["income"]), 0)
        self.assertEqual(len(self.testBook.book["expense"]), 0)

    def test_displayBook(self):
        
        income_transaction = Income(500.0, "Displayable Income", "salary income")
        expense_transaction = Expense(75.0, "Displayable Expense", "entertainment")
        self.testBook.addTran(income_transaction, self.filePath)
        self.testBook.addTran(expense_transaction, self.filePath)

        captured_output = StringIO()
        sys.stdout = captured_output

        self.testBook.displayBook()

        sys.stdout = sys.__stdout__

        output = captured_output.getvalue()
        self.assertIn("Displayable Income", output)
        self.assertIn("Displayable Expense", output)

unittest.main(argv=[''], verbosity=2, exit=False)




from datetime import datetime
from datetime import date
from .negative_amount_error import NegativeAmountError

class Transaction:
    """
    This class represents Transaction object
    """


    def __init__(self, amount, desc, label=None):
        try:
            self.amount = float(amount)
            if self.amount < 0:
                raise NegativeAmountError              
        except ValueError:
            print("Amount must be a number")
        except NegativeAmountError as e:
            print(f"Error: {e}")
                   

        self.desc = desc
        self.date = date.today()
        self.label = label if label else "other"             

    def setDate(self, newDate):
        try:
            if isinstance(newDate, str):
                self.date = datetime.strptime(str(newDate), "%Y-%m-%d").date()
            else:
                print("The new date must be a string")
        except ValueError:
            print("Invalid date format. Please use 'YYYY-MM-DD'.")

    def setLabel(self, newLabel):
        try:
            if isinstance(newLabel, str):
                self.label = newLabel
            else:
                raise ValueError("The new Label must be a string")
        except ValueError as ve:
            print(ve) 

    def setAmount(self, newAmount):
        try:
            # Attempt to convert the input to a float
            float_value = float(newAmount)

            # Check if the float conversion was successful
            if isinstance(float_value, float):
                self.amount = float_value
            else:
                raise ValueError("The new amount must be a number")
        except ValueError as ve:
            print(ve)     

    def toDict(self):
        return {
            'amount': self.amount,
            'date': self.date.strftime('%Y-%m-%d'),
            'desc': self.desc,
            'label': self.label
        }

    def __str__(self):
        return f"{{'amount': {self.amount}, 'date': '{self.date.strftime('%Y-%m-%d')}', 'description': '{self.desc}', 'label': '{self.label}'}}"
    
class Income(Transaction):
    """
    A subclass of Transaction class that represents Income object
    """

    incomeLabels = ["salary income", "passive income", "capital gains", "other"]

    def __init__(self, amount, desc, label=None):
        super().__init__(amount, desc, label)
        if label:
            if label not in Income.incomeLabels:
                Income.incomeLabels.append(label)

class Expense(Transaction):
    """
    A subclass of Transaction class that represents Expense obejct
    """

    expenseLabels = ["utility", "grocery", "transportation", "entertainment", "healthcare", "other"]

    def __init__(self, amount, desc, label=None):
        super().__init__(amount, desc, label)
        if label:
            if label not in Expense.expenseLabels:
                Expense.expenseLabels.append(label)


class NegativeAmountError(Exception):
    """User-defined exception for handling negative amount."""

    def __init__(self, message = "Input amount should be greater than or equal to zero."):
            
        self.message = message
        super().__init__(self.message)


import unittest
from ..src.bookkeeping.management.transaction import Income
from ..src.bookkeeping.management.transaction import Expense
from datetime import datetime
from datetime import date

class TestTransaction(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        pass

    def setUp(self):
        # Common setup for each test case
        self.income_1 = Income(100.0, "Test Income Transaction 1", "salary income")
        self.income_2 = Income(200.0, "Test Income Transaction 2", "passive income")     
        self.expense_1 = Expense(300.0, "Test Expense Transaction 1", "grocery")
        self.expense_2 = Expense(400.0, "Test Expense Transaction 2", "entertainment")

    def tearDown(self):
        del self.income_1
        del self.income_2
        del self.expense_1
        del self.expense_2

    def test_setDate(self):
        self.income_1.setDate('2014-01-01')
        self.income_2.setDate('2014-01-02')
        self.expense_1.setDate('2014-01-03')
        self.expense_2.setDate('2014-01-04')

        self.assertEqual(self.income_1.date, datetime.strptime(str("2014-01-01"), "%Y-%m-%d").date())
        
        self.assertEqual(self.income_2.date, datetime.strptime(str("2014-01-02"), "%Y-%m-%d").date())
        self.assertEqual(self.expense_1.date, datetime.strptime(str("2014-01-03"), "%Y-%m-%d").date())
        self.assertEqual(self.expense_2.date, datetime.strptime(str("2014-01-04"), "%Y-%m-%d").date())

    def test_setLabel_valid(self):
        self.income_1.setLabel("passive income")
        self.income_2.setLabel("other")
        self.expense_1.setLabel("utility")
        self.expense_2.setLabel("newlabel")

        self.assertEqual(self.income_1.label, "passive income")
        self.assertEqual(self.income_2.label,"other")
        self.assertEqual(self.expense_1.label, "utility")
        self.assertEqual(self.expense_2.label, "newlabel")

    def test_setAmount_valid(self):
        self.income_1.setAmount(1000)
        self.income_2.setAmount(2000)
        self.expense_1.setAmount(3000)
        self.expense_2.setAmount(4000)

        self.assertEqual(self.income_1.amount, 1000)
        self.assertEqual(self.income_2.amount, 2000)
        self.assertEqual(self.expense_1.amount, 3000)
        self.assertEqual(self.expense_2.amount, 4000)

    def test_toDict(self):
        # Test the toDict method
        expected_dict_in_1 = {
            'amount': 100.0,
            'date': date.today().strftime('%Y-%m-%d'),
            'desc': "Test Income Transaction 1",
            'label': "salary income"
        }
        expected_dict_in_2 = {
            'amount': 200.0,
            'date': date.today().strftime('%Y-%m-%d'),
            'desc': 'Test Income Transaction 2',
            'label': "passive income"
        }

        expected_dict_ep_1 = {
            'amount': 300.0,
            'date': date.today().strftime('%Y-%m-%d'),
            'desc': 'Test Expense Transaction 1',
            'label': "grocery"
        }
        expected_dict_ep_2 = {
            'amount': 400.0,
            'date': date.today().strftime('%Y-%m-%d'),
            'desc': "Test Expense Transaction 2",
            'label': "entertainment"
        }
        result_dict_in_1 = self.income_1.toDict()
        result_dict_in_2= self.income_2.toDict()
        result_dict_ep_1= self.expense_1.toDict()
        result_dict_ep_2= self.expense_2.toDict()

        self.assertEqual(result_dict_in_1, expected_dict_in_1)
        self.assertEqual(result_dict_in_2, expected_dict_in_2)
        self.assertEqual(result_dict_ep_1, expected_dict_ep_1)
        self.assertEqual(result_dict_ep_2, expected_dict_ep_2)

    def test_str(self):
        expected_str_in_1 = f"{{'amount': 100.0, 'date': '{date.today().strftime('%Y-%m-%d')}', 'description': 'Test Income Transaction 1', 'label': 'salary income'}}"
        expected_str_in_2 = f"{{'amount': 200.0, 'date': '{date.today().strftime('%Y-%m-%d')}', 'description': 'Test Income Transaction 2', 'label': 'passive income'}}"
        expected_str_ep_1 = f"{{'amount': 300.0, 'date': '{date.today().strftime('%Y-%m-%d')}', 'description': 'Test Expense Transaction 1', 'label': 'grocery'}}"
        expected_str_ep_2 = f"{{'amount': 400.0, 'date': '{date.today().strftime('%Y-%m-%d')}', 'description': 'Test Expense Transaction 2', 'label': 'entertainment'}}"
  
        
        result_str_in_1 = self.income_1.__str__()
        result_str_in_2 = self.income_2.__str__()
        result_str_ep_1 = self.expense_1.__str__()
        result_str_ep_2 = self.expense_2.__str__()

        self.assertEqual(result_str_in_1, expected_str_in_1)
        self.assertEqual(result_str_in_2, expected_str_in_2)
        self.assertEqual(result_str_ep_1, expected_str_ep_1)
        self.assertEqual(result_str_ep_2, expected_str_ep_2)

    @classmethod
    def tearDownClass(cls):
        pass

unittest.main(argv=[''], verbosity=2, exit=False)


from datetime import datetime
import pygal
from pygal.style import DefaultStyle
from IPython.display import SVG, display
from .analytics import Analysis


class SummaryAnalysis(Analysis):
    """
    Implementation of a SummaryAnalysis class which contains methods
    to perform basic analysis on monthly transactions
    """

    def __init__(self, file_path = "userBook.json"):
        super().__init__(file_path)

    def checkBalance(self):
        """ 
        Calculate overall balance for the current month and print the result 
        """
        current_month = datetime.now().strftime('%Y-%m')
        current_month_transactions = [
            t for t in self.transactions if t['date'][:7] == current_month]
        current_month_balance = sum(
            t['amount'] if t['type'] == 'income' else -t['amount'] for t in current_month_transactions)
        print(f"Overall Balance for {current_month}: {current_month_balance}")

    def displayTrend(self, graph_type):
        """ 
        Plot a line chart or bar chart to represent the monthly 
        trend of income, expenses and balance

        Args:
            graph_type (int): 1 for line chart, 2 for bar chart

        Raises:
            ValueError: invalid argument input for graph_type
        """
        # Calculate monthly totals
        monthly_totals = {}
        for transaction in self.transactions:
            transaction_date = datetime.strptime(
                transaction['date'], '%Y-%m-%d')
            month_key = transaction_date.strftime('%Y-%m')
            if month_key not in monthly_totals:
                monthly_totals[month_key] = {
                    'income': 0, 'expense': 0, 'balance': 0}
            if transaction['type'] == 'income':
                monthly_totals[month_key]['income'] += transaction['amount']
            elif transaction['type'] == 'expense':
                monthly_totals[month_key]['expense'] += transaction['amount']

        # Calculate balance for each month
        for month_key, totals in monthly_totals.items():
            totals['balance'] = totals['income'] - totals['expense']

        # Extract data for plotting
        sorted_months = sorted(monthly_totals.keys())
        income_values = [monthly_totals[month]['income']
                         for month in sorted_months]
        expense_values = [monthly_totals[month]['expense']
                          for month in sorted_months]
        balance_values = [monthly_totals[month]['balance']
                          for month in sorted_months]

        # Plot line chart or bar chart
        chart = None
        try: 
            graph_type = int(graph_type)
            if graph_type == 1:  # line chart
                chart = pygal.Line(print_values=True,
                                    style=DefaultStyle(value_font_size=10))
            elif graph_type == 2:  # bar chart
                chart = pygal.Bar(print_values=True, print_values_position='top',
                                    style=DefaultStyle(value_font_size=10))
            else:
                raise ValueError(
                    "Please use pre-defined type code for graph_type")
            chart.title = 'Monthly trend of expenses, income and balance'
            chart.x_title = 'month'
            chart.y_title = 'amount'
            chart.x_labels = map(str, sorted_months)
            chart.add('Expenses', expense_values)
            chart.add('Income', income_values)
            chart.add('Balance', balance_values)
            display(SVG(chart.render(disable_xml_declaration=True)))
        except ValueError as v:
            print(f"Value Error: {v}")


    def displayByLabel(self, transaction_type, graph_type):
        """ 
        Plot one type of pie chart to represent the constitution
        of income or expenses

        Args:
            transaction_type (str): "income", "expense"
            graph_type (int): 1 for pie chart, 2 for donut chart, 
                            3 for ring chart, 4 for half-pie chart

        Raises:
            ValueError: invalid argument input for transaction_type
            ValueError: invalid argument input for graph_type
        """
        # Calculate totals by label
        label_totals = {'income': {}, 'expense': {}}
        for transaction in self.transactions:
            label_type = transaction['type']
            label = transaction['label']
            if label not in label_totals[label_type]:
                label_totals[label_type][label] = 0
            label_totals[label_type][label] += transaction['amount']

        # Store label composition
        percentage_composition = {
            'income': label_totals['income'], 'expense': label_totals['expense']}
        total_income = sum(label_totals['income'].values())
        for label, amount in label_totals['income'].items():
            percentage_composition['income'][label] = (
                amount / total_income) * 100
        total_expense = sum(label_totals['expense'].values())
        for label, amount in label_totals['expense'].items():
            percentage_composition['expense'][label] = (
                amount / total_expense) * 100

        # Plot a type of pie chart of income or expenses
        pie_chart = None
        try:
            graph_type = int(graph_type)
            if graph_type == 1:  # Pie
                pie_chart = pygal.Pie(print_values=True, style=DefaultStyle(
                    value_font_size=10), value_formatter=lambda x:  '%.2f%%' % float(x))
            elif graph_type == 2:  # Donut
                pie_chart = pygal.Pie(inner_radius=.4, print_values=True, style=DefaultStyle(
                    value_font_size=10), value_formatter=lambda x:  '%.2f%%' % float(x))
            elif graph_type == 3:  # Ring
                pie_chart = pygal.Pie(inner_radius=.75, print_values=True, style=DefaultStyle(
                    value_font_size=10), value_formatter=lambda x:  '%.2f%%' % float(x))
            elif graph_type == 4:  # Half-pie
                pie_chart = pygal.Pie(half_pie=True, print_values=True, style=DefaultStyle(
                    value_font_size=10), value_formatter=lambda x:  '%.2f%%' % float(x))
            else:
                raise ValueError(
                    "Please use pre-defined type code for graph_type")

            transaction_type = str(transaction_type)
            if transaction_type == "income":
                pie_chart.title = 'Constitution of income'
                for key, value in percentage_composition['income'].items():
                    pie_chart.add(key, value)
                display(SVG(pie_chart.render(disable_xml_declaration=True)))
            elif transaction_type == "expense":
                pie_chart.title = 'Constitution of expenses'
                for key, value in percentage_composition['expense'].items():
                    pie_chart.add(key, value)
                display(SVG(pie_chart.render(disable_xml_declaration=True)))
            else:
                raise ValueError(
                    "Please use pre-defined type code for transaction_type")
        except ValueError as v:
            print(f"Value Error: {v}")



import json
from .transaction import Income
from .transaction import Expense

class Book:
    """
    This class represents the Book object that keeps records of 
    all transactions of users
    """

    def __init__(self, filePath = "userBook.json"):
        self.book = {"expense": [],
                     "income": []}
        self.loadData(filePath)

    def saveData(self, filePath = "userBook.json"):
        with open(filePath, 'w') as file:
            json.dump(self.book, file, indent=2)

    def loadData(self, filePath):
        try:
            with open(filePath, 'r') as file:
                self.book = json.load(file)
        except (FileNotFoundError, json.JSONDecodeError):
            self.book = {"expense": [],
                         "income": []}

    def addTran(self, transaction, filePath = "userBook.json"):
        try:
            transactionDict = transaction.toDict()
            if isinstance(transaction, Income):
                self.book["income"].append(transactionDict)
            elif isinstance(transaction, Expense):
                self.book["expense"].append(transactionDict)
            self.saveData(filePath)
            print(f"The below transaction has been added to the book: \n{transactionDict}")

        except Exception as e:
            print(f"Error adding transaction: {e}")

    def searchTran(self, t_type, keyword):
        found = False
        try:
            if not isinstance(t_type, str) or (t_type.lower() not in ["income", "expense"]):
                raise ValueError("Transaction type must be a string ('income' or 'expense').")
            if not isinstance(keyword, str):
                raise ValueError("Keyword must be a string.")
            if t_type in self.book:
                
                for index, item in enumerate(self.book[t_type]):
                        if keyword in item["desc"]:
                            index += 1
                            print(index, item)
                            found = True
                if not found:
                        print(f"{keyword} not found in the {t_type} transaction records.")
                        return False
            else:
                print(f"{t_type} category not found in the book.")
                return False
        except ValueError as ve:
            print(f"Error searching transaction: {ve}")
            return False
        finally:
            return found
    
    def removeTran(self, t_type, index, filePath = "userBook.json"):
        try:
            if not isinstance(t_type, str) or t_type.lower() not in ["income", "expense"]:
                raise ValueError("Transaction type must be a string ('income' or 'expense').")
            index = int(index) - 1
            if not (0 <= index < len(self.book[t_type])):
                raise ValueError("index is out of bounds.")
            
            if t_type in self.book:
                print(f"Removing {self.book[t_type][index]}")
                del self.book[t_type][index]
                print(f"transaction {index + 1} in {t_type} has been removed")
                self.saveData(filePath)
            else:
                print(f"Error: {t_type} category not found in the book.")
        except ValueError as e:
            print(f"Error removing transaction: {e}")

    def clearAll(self, filePath = "userBook.json"):
        self.book["income"] = []
        self.book["expense"] = []
        self.saveData(filePath)
        print("All transactions have been clear in this book")        

    def displayBook(self):
        print(f"All transactions:")
        for t_type in ['income', 'expense']:
            index = 1
            print(f"{t_type}:")
            for transaction in self.book[t_type]:
                print(f"{index}. {transaction}")
                index += 1


import unittest
from unittest.mock import patch
from ..src.bookkeeping.analysis.budget import BudgetAnalysis

class TestBudgetAnalysis(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        print("Setting up the test class")

    @classmethod
    def tearDownClass(cls):
        print("Tearing down the test class")

    def setUp(self):
        print("Setting up for a test")
        self.analysis_instance = BudgetAnalysis(file_path = "userBook.json")

    def tearDown(self):
        self.analysis_instance = None
        print("Tearing down after a test")

    def test_setBudget(self):
        # Test setting the budget with a number (type = 1)
        with patch("builtins.print") as mock_print:
            self.analysis_instance.setBudget(1, 500)
            self.assertEqual(self.analysis_instance.budget, 500)
            mock_print.assert_called_once_with("Your set monthly budget is 500 CAD")

        # Test setting the budget with a percentage (type = 2) and original data
        with patch("builtins.print") as mock_print:
            self.analysis_instance.setBudget(2, 20)
            self.assertAlmostEqual(self.analysis_instance.budget, 157, places = 0)
            mock_print.assert_called_once_with("Your set monthly budget is 157 CAD")        


    def test_overBudgetExpenses(self):
        # Test the overBudgetExpenses method with a higher budget
        self.analysis_instance.budget = 500
        with patch("builtins.print") as mock_print:
            self.analysis_instance.overBudgetExpenses()
            mock_print.assert_called_once_with("Expenses in the current month have not exceed the budget, 337 CAD remaining")

        # Test the overBudgetExpenses method with a lower budget
        self.analysis_instance.budget = 100
        with patch.object(self.analysis_instance, 'displayTransactions') as mock_display:
            # Mock the print function
            with patch("builtins.print") as mock_print:
                self.analysis_instance.overBudgetExpenses()

        # Check if the expected output is present in the printed output
        expected_output = "Expenses in the current month have exceeded the budget by 63 CAD"
        mock_print.assert_called_with(expected_output)

        # Check if the 'displayTransactions' method was called
        mock_display.assert_called_once()

        # Check the displayed transactions
        expected_transactions = [
            {'type': 'expense', 'amount': 122.0, 'date': '2023-12-01', 'description': 'description', 'label': 'utility'}
            ]        
        mock_display.assert_called_with(expected_transactions)      

if __name__ == '__main__':
    unittest.main(argv =[''], verbosity=2, exit=False)


import json

class Analysis:
    """
    Implementation of a Analysis class which loads transaction data
    from a JSON file and stores in an attribute
    """

    def __init__(self,file_path = "userBook.json"):
        self.transactions = self.loadBookData(file_path)

    def loadBookData(self, file_path):
        # Load data from the JSON file
        # The default path for JSON file is 'bookkeeping/userBook.json'
        try:
            with open(file_path, 'r') as file:
                data = json.load(file)
        except FileNotFoundError:
            # Handle file not found error
            print(f"Error: File '{file_path}' not found.")
            return []


        # Extract and combine transaction information
        transactions = []
        try:
            for expense in data.get('expense', []):
                transaction_info = {
                    'type': 'expense',
                    'amount': float(expense.get('amount')),
                    'date': expense.get('date'),
                    'description': expense.get('desc'),
                    'label': expense.get('label')
                }
                transactions.append(transaction_info)

            for income in data.get('income', []):
                transaction_info = {
                    'type': 'income',
                    'amount': float(income.get('amount')),
                    'date': income.get('date'),
                    'description': income.get('desc'),
                    'label': income.get('label')
                }
                transactions.append(transaction_info)
        except (AttributeError, ValueError) as e:
            # Handle attribute or value errors during data extraction
            print(f"Error: Unable to extract data from file '{file_path}'. {e}")
            return []

        return transactions


from datetime import datetime
import pandas as pd
import pygal
from pygal.style import DefaultStyle
from IPython.display import SVG, display
from .analytics import Analysis
from ..management.negative_amount_error import NegativeAmountError

class BudgetAnalysis(Analysis):
    """
    Implementation of a BudgetAnalysis class which contains methods
    to perform basic analysis that is related to budget
    """

    def __init__(self, file_path = "userBook.json"):
        super().__init__(file_path)
        self.budget = 0  # Initialize budget as 0

    def setBudget(self, type, amount):
        """
        Set a monthly budget

        Args:
            type (int): an integer indicating whether the user would like to 
                    set the budget with the amount number or percentage of income
                    1 for an amount of monthly budget, 2 for a percentage of average income
            amount (float): an amount(>0) or a percentage(0-100)

        Raises:
            ValueError: invalid argument input
        """
        try:
            type = int(type)
            if type == 1:
                self.budget = float(amount)
            elif type == 2:
                if amount < 0:
                    raise NegativeAmountError
                percentage = float(amount)/100
                average_income = sum(t['amount'] for t in self.transactions if t['type']
                                        == 'income') / len(set(t['date'][:7] for t in self.transactions))
                self.budget = percentage * average_income
            else:
                raise ValueError("Please use pre-defined type code")
            print(f"Your set monthly budget is {round(self.budget)} CAD")
        except ValueError as v:
            print(f"Value Error: {v}")
        except NegativeAmountError as n:
            print(f"Negative Amount Error: {n}")
   

    def overBudgetExpenses(self):
        """ 
        Print the amount of the expenses exceeding budget of this month 
        and all expenses of this month that occur after the point where
        the sum of expenses exceeds set budget
        """
        current_month = datetime.now().strftime('%Y-%m')
        current_month_expense = sum(
            t['amount'] for t in self.transactions if t['date'][:7] == current_month and t['type'] == 'expense')
        if self.budget != 0:
            over_budget = self.budget-current_month_expense
            if over_budget >= 0:
                print(
                    f"Expenses in the current month have not exceed the budget, {round(over_budget)} CAD remaining")
            else:
                print(
                    f"Expenses in the current month have exceeded the budget by {-round(over_budget)} CAD")
        else:
            raise ValueError("Budget has not been set yet")

        post_budget_transactions = self.transactionsAfterBudgetExceeded()
        self.displayTransactions(post_budget_transactions)

    def transactionsAfterBudgetExceeded(self):
        """
        Get all expenses of this month that occur after the point where
        the sum of expenses exceeds set budget

        Returns:
            post_budget_transactions(list): a list of transactions 
        """
        total_expense = 0
        budget_exceeded = False
        post_budget_transactions = []
        current_month = datetime.now().strftime('%Y-%m')

        for transaction in self.transactions:
            if transaction['type'] == 'expense':
                month = transaction['date'][:7]
                if month != current_month:
                    continue
                total_expense += transaction['amount']
                if total_expense > self.budget and not budget_exceeded:
                    budget_exceeded = True
                if budget_exceeded:
                    post_budget_transactions.append(transaction)

            return post_budget_transactions

    def displayTransactions(self, transactions):
        """
        Store the passed-in transactions as a pandas dataframe
        and print them
        """
        # Continue only if there is any transaction to show
        if transactions:
            # Create a DataFrame from the transactions list
            df = pd.DataFrame(transactions)

            # Format the DataFrame for better readability
            df['amount'] = df['amount'].apply(lambda x: f"${x:,.2f}")
            df['date'] = pd.to_datetime(df['date']).dt.strftime('%Y-%m-%d')

            # Sort by date for chronological order
            df.sort_values(by='date', inplace=True)

            # Reset index for neatness
            df.reset_index(drop=True, inplace=True)

            print(df)

    def displayOverBudgetTrend(self):
        """ 
        Plot a bar chart to show the monthly trend of expenses exceeding over budget"""
        # Calculate monthly expenses
        budget_monthly_totals = {}
        for transaction in self.transactions:
            transaction_date = datetime.strptime(
                transaction['date'], '%Y-%m-%d')
            month_key = transaction_date.strftime('%Y-%m')
            if month_key not in budget_monthly_totals:
                budget_monthly_totals[month_key] = {
                    'expense': 0, 'overBudget': 0}
            if transaction['type'] == 'expense':
                budget_monthly_totals[month_key]['expense'] += transaction['amount']

        # Calculate the amount of expense over budget for each month
        for month_key, totals in budget_monthly_totals.items():
            totals['overBudget'] = totals['expense'] - self.budget

        # Extract data for plotting
        sorted_months = sorted(budget_monthly_totals.keys())
        overbudget_values = [budget_monthly_totals[month]
                                ['overBudget'] for month in sorted_months]

        # plot a bar chart for trend of expenses over budget
        bar_chart = pygal.Bar(print_values=True, print_values_position='top',
                                value_formatter=lambda x:  '%d' % x,
                                style=DefaultStyle(value_font_size=10))
        bar_chart.title = 'Monthly trend of expenses over budget'
        bar_chart.x_title = 'month'
        bar_chart.y_title = 'amount'
        bar_chart.x_labels = map(str, sorted_months)
        bar_chart.add('overBudget', overbudget_values)
        display(SVG(bar_chart.render(disable_xml_declaration=True)))
